"""Analyze report files, and print results.
examples:
analyze(["a.csv", "b.csv"], 30)
analyze("a.csv", 30)

Output:
Case Id,Fail Count,Total Count,Failure Rate,Minimum Time Taken,Maximum Time Taken,Average Time Taken
case1,136,7670,1.7731%,3.69,26.583,16.610214623718797
case2,0,4,0.0000%,16.461,26.072,20.065125


Case Id,Group Index,Start Time,End Time,Fail Count,Total Count,Failure Rate,Minimum Time Taken,Maximum Time Taken,Average Time Taken
case1,1,2018-06-18 10:32:00,2018-06-18 11:02:00,30,1938,1.5480%,3.83,26.583,16.469082024050984
case1,2,2018-06-18 11:02:00,2018-06-18 11:32:00,42,2000,2.1000%,3.72,18.932,16.959928979894286
case1,3,2018-06-18 11:32:00,2018-06-18 12:02:00,38,1932,1.9669%,3.69,25.363,16.23456723705196
case1,4,2018-06-18 12:02:00,2018-06-18 12:32:00,26,1798,1.4461%,3.72,20.412,16.72942924743759
case1,5,2018-06-18 12:32:00,2018-06-18 13:02:00,0,2,0.0000%,16.491,16.491,16.491
case2,1,2018-06-18 10:32:00,2018-06-18 11:02:00,0,2,0.0000%,26.072,26.072,26.072
case2,2,2018-06-18 11:02:00,2018-06-18 11:32:00,0,0,0.0000%,,,
case2,3,2018-06-18 11:32:00,2018-06-18 12:02:00,0,2,0.0000%,16.461,16.461,16.461
"""
import datetime
import os
import re

from eztest import utility

try:
    from _collections import OrderedDict
except ImportError:
    from collections import OrderedDict


AVERAGE = 'average'
FAIL_COUNT = 'fail_count'
ID = 'id'
MAX_TIME = 'max_time'
MIN_TIME = 'min_time'
PASS_COUNT = 'pass_count'
START_TIME = 'start_time'
STATUS_PATTERN = re.compile(r'^"\d+","(.+?)",".+?","(Pass|Fail)"')
TIME_PATTERN = re.compile(r'"(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6})","(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2}\.\d{6})","([\d\\.]+)"')
TOTAL_COUNT = 'total_count'


def _add_to_group_summary(summary, id, start_time, time_taken=None, is_case_pass=None):
    """Add to group summary.

    :param dict summary: group summary.
    :param str id: case id.
    :param datetime.datetime start_time: start datetime.
    :param float time_taken: time taken in second.
    :param bool is_case_pass: is case pass.
    """
    key = "{}_{}".format(id, utility.date2str(start_time, '%Y%m%d%H%M%S'))
    if time_taken is None:
        summary[key] = {ID: id,
                        START_TIME: start_time}
    elif key not in summary:
        summary[key] = {ID: id,
                        START_TIME: start_time,
                        AVERAGE: time_taken,
                        TOTAL_COUNT: 1,
                        FAIL_COUNT: 0 if is_case_pass else 1,
                        MIN_TIME: time_taken,
                        MAX_TIME: time_taken
                        }
    else:
        summary[key][TOTAL_COUNT] +=1
        if not is_case_pass:
            summary[key][FAIL_COUNT] += 1
        summary[key][AVERAGE] = (summary[key][AVERAGE] + time_taken) / 2.0
        summary[key][MIN_TIME] = min(summary[key][MIN_TIME], time_taken)
        summary[key][MAX_TIME] = max(summary[key][MAX_TIME], time_taken)


def _add_to_case_summary(summary, id, time_taken, is_case_pass):
    """Add to case summary.

    :param dict summary: case summary.
    :param str id: case id.
    :param float time_taken: time taken in second.
    :param bool is_case_pass: is case pass.
    """
    if id not in summary:
        summary[id] = {AVERAGE: time_taken,
                       TOTAL_COUNT: 1,
                       FAIL_COUNT: 0 if is_case_pass else 1,
                       MIN_TIME: time_taken,
                       MAX_TIME: time_taken}
    else:
        summary[id][TOTAL_COUNT] += 1
        if not is_case_pass:
            summary[id][FAIL_COUNT] += 1
        summary[id][AVERAGE] = (summary[id][AVERAGE] + time_taken) / 2.0
        summary[id][MIN_TIME] = min(summary[id][MIN_TIME], time_taken)
        summary[id][MAX_TIME] = max(summary[id][MAX_TIME], time_taken)


def analyze(file_paths, group_minutes=60):
    """Analyze report files and calculate failure rate, average of time taken.

    :param list|str file_paths: file paths.
    :param int group_minutes: calculate failure rate and average of time taken by grouping case results with [group_minutes] minutes.
    """
    if not file_paths:
        raise ValueError('Plese provide file path.')
    group_gap = datetime.timedelta(seconds=group_minutes * 60)
    if not isinstance(file_paths, list):
        file_paths = [file_paths]

    group_summary = OrderedDict()
    case_summary = dict()
    for file_path in file_paths:
        if not os.path.exists(file_path):
            print('Error: cannot find {}.'.format(file_path))
            continue

        print('Calculating for {}...'.format(file_path))
        start_times, id, is_pass = dict(), None, None
        with open(file_path, 'r') as f:
            while True:
                line = f.readline()
                if not line:
                    break
                status_match = STATUS_PATTERN.match(line)
                time_match = TIME_PATTERN.search(line)
                if status_match:
                    id = status_match[1]
                    is_pass = True if status_match[2] == 'Pass' else False

                if time_match:
                    start_date, end_date, time_taken = utility.str2date(time_match[1]), utility.str2date(time_match[2]), float(time_match[3])
                    _add_to_case_summary(case_summary, id, time_taken, is_pass)

                    if id not in start_times:
                        start_times[id] = start_date.replace(second=0, microsecond=0)
                    my_start_time = start_times[id]

                    if my_start_time + group_gap >= end_date:
                        _add_to_group_summary(group_summary, id, my_start_time, time_taken, is_pass)
                    else:
                        while True:
                            my_start_time += group_gap
                            start_times[id] = my_start_time
                            if my_start_time + group_gap < end_date:
                                _add_to_group_summary(group_summary, id, my_start_time)
                            else:
                                break
                        _add_to_group_summary(group_summary, id, my_start_time, time_taken, is_pass)

    if not group_summary:
        print('Invalid report file')
    else:
        print("\n\nCase Id,Fail Count,Total Count,Failure Rate,Minimum Time Taken,Maximum Time Taken,Average Time Taken")
        groups = []
        for id, value in case_summary.items():
            print("{},{},{},{},{},{},{}".format(
                id,
                value[FAIL_COUNT],
                value[TOTAL_COUNT],
                '{:.4f}%'.format(value[FAIL_COUNT]/(1 if value[TOTAL_COUNT] == 0 else value[TOTAL_COUNT]) * 100),
                value[MIN_TIME],
                value[MAX_TIME],
                value[AVERAGE]
            ))
            index = 1
            for key, group_value in group_summary.items():
                if key.startswith(id):
                    groups.append("{},{},{},{},{},{},{},{},{},{}".format(
                        id,
                        index,
                        utility.date2str(group_value[START_TIME], '%Y-%m-%d %H:%M:%S'),
                        utility.date2str(group_value[START_TIME] + group_gap, '%Y-%m-%d %H:%M:%S'),
                        group_value.get(FAIL_COUNT, 0),
                        group_value.get(TOTAL_COUNT, 0),
                        '{:.4f}%'.format(group_value[FAIL_COUNT] / (1 if group_value[TOTAL_COUNT] == 0 else group_value[TOTAL_COUNT]) * 100
                                         if FAIL_COUNT in group_value else 0),
                        group_value.get(MIN_TIME, ''),
                        group_value.get(MAX_TIME, ''),
                        group_value.get(AVERAGE, '')
                    ))
                    index += 1

        print('\n\nCase Id,Group Index,Start Time,End Time,Fail Count,Total Count,Failure Rate,Minimum Time Taken,Maximum Time Taken,Average Time Taken')
        for group in groups:
            print(group)
